from __future__ import absolute_import, unicode_literals, print_function
import json
import os

import logging
logger = logging.getLogger(__name__)

class Option(object):
    def __init__(self, name, helptext):
        self.name = name
        self.helptext = helptext

class CommandTree(object):
    def __init__(self, node=None, helptext=None, children=None, localFlags=None):
        self.node = node
        self.help = helptext
        self.children = children if children else list()
        self.localFlags = localFlags if localFlags else list()

    def __str__(self):
        return "Node: %s, Help: %s\n Flags: %s\n Children: %s" % (self.node, self.help, self.localFlags, self.children)

class Parser(object):
    def __init__(self, apiFile, root_name):
        try:
            self.json_api = apiFile
            self.schema = dict()
            self.globalFlags = list()
            self.root_name = root_name
            with open(self.json_api) as api:
                self.schema = json.load(api)
            self.ast = CommandTree(root_name)  # Set to dynamic root name
            root_schema = self.schema.get(root_name)
            if root_schema is None:
                logger.error(f"{root_name} is not in the schema")
            else:
                self.ast = self.build(self.ast, root_schema)
        except Exception as ex:
            logger.error(f"Exception while initializing Parser: {ex}")

    def set_api_file(self, apiFile):
        try:
            self.json_api = apiFile
            logger.info(f"Setting API file to {self.json_api}")
            with open(self.json_api) as api:
                self.schema = json.load(api)
            # We assume the top-level key in each JSON is the name of the API, like "oci" or "aws"
            api_schema = self.schema.get(self.json_api.split('/')[-1].split('.')[0])
            if api_schema is None:
                logger.error(f"{self.json_api.split('/')[-1].split('.')[0]} is not in the schema")
            else:
                self.ast = self.build(self.ast, api_schema)
        except Exception as ex:
            logger.error(f"Exception while setting API file: {ex}")

    def get_top_level_commands(self):
        return [node.node for node in self.ast.children]

    def build(self, root, schema):
        try:
            if schema.get("subcommands") and schema["subcommands"]:
                for subcmd, childSchema in schema["subcommands"].items():
                    child = CommandTree(node=subcmd)
                    child = self.build(child, childSchema)
                    root.children.append(child)
            # Add options and arguments to the current node (root)
            try:
                for name, desc in schema.get("options", {}).items():
                    root.localFlags.append(Option(name, desc["help"]))
            except:
                pass
            try:
                for arg in schema.get("args", []):
                    node = CommandTree(node=arg)
                    root.children.append(node)
            except:
                pass
            root.help = schema.get("help")
            return root
        except:
            pass

    def print_tree(self, root, indent=0):
        indentter = '{:>{width}}'.format(root.node, width=indent)
        print(indentter)
        for child in root.children:
            self.print_tree(root=child, indent=indent+2)

    def parse_tokens(self, tokens):
        if len(tokens) == 1:
            return list(), tokens, {"oci": self.ast.help}
        else:
            tokens.reverse()
    
        parsed, unparsed, _ = self.treewalk(self.ast, parsed=list(), unparsed=tokens)
    
        # Use parsed tokens to find current node in command tree
        current_node = self.get_node_from_tokens(parsed)
    
        # Populate suggestions based on the children of the current node
        suggestions = {child.node: child.help for child in current_node.children}
    
        # Also add options available at the current node to suggestions
        suggestions.update({flag.name: flag.helptext for flag in current_node.localFlags})
    
        if not suggestions and unparsed:
            logger.debug("unparsed tokens remain, possible value encountered")
            unparsed.pop()
            parsed.reverse()
            unparsed.extend(parsed)
            logger.debug("resuming treewalk with tokens: %s", unparsed)
            return self.treewalk(self.ast, parsed=list(), unparsed=unparsed)
        else:
            return parsed, unparsed, suggestions
    
    def get_node_from_tokens(self, parsed_tokens):
        current_node = self.ast
        for token in parsed_tokens:
            if token.startswith("--"):  # Skip options
                continue
            for child in current_node.children:
                if child.node == token:
                    current_node = child
                    break
        return current_node
    
    def treewalk(self, root, parsed, unparsed):
        suggestions = dict()
        if not unparsed:
            logger.debug("no tokens left unparsed. returning %s, %s", parsed, suggestions)
            return parsed, unparsed, suggestions
        token = unparsed.pop().strip()
        logger.debug("begin parsing at %s w/ tokens: %s", root.node, unparsed)

        if root.node == token:
            logger.debug("root node: %s matches next token:%s", root.node, token)
            parsed.append(token)

            if self.peekForOption(unparsed):  # check for localFlags and globalFlags
                logger.debug("option(s) upcoming %s", unparsed)
                parsed_opts, unparsed, suggestions = self.evalOptions(root, list(), unparsed[:])
                if parsed_opts:
                    logger.debug("parsed option(s): %s", parsed_opts)
                    parsed.extend(parsed_opts)

            if unparsed and not self.peekForOption(unparsed):  # unparsed bits without options
                logger.debug("begin subtree %s parsing", root.node)
                for child in root.children:
                    parsed_subtree, unparsed, child_suggestions = self.treewalk(child, list(), unparsed[:])
                    if parsed_subtree:  # subtree returned further parsed tokens
                        parsed.extend(parsed_subtree)
                        logger.debug("subtree at: %s has matches. %s, %s", child.node, parsed, unparsed)
                        suggestions.update(child_suggestions)
                        break
                else:
                    logger.debug("no matches in subtree: %s. returning children as suggestions", root.node)
                    for child in root.children:
                        suggestions[child.node] = child.help
        else:
            logger.debug("no token or option match. Current node: %s, Token: %s", root.node, token)
            unparsed.append(token)

        logger.debug("current state - parsed: %s, unparsed: %s, suggestions: %s", parsed, unparsed, suggestions)
        return parsed, unparsed, suggestions

    def peekForOption(self, unparsed):
        if unparsed and (unparsed[-1].startswith("--") or unparsed[-1].startswith("-")):
            return True
        return False

    def evalOptions(self, root, parsed, unparsed):
        logger.debug("parsing options at tree: %s with p:%s, u:%s", root.node, parsed, unparsed)
        suggestions = dict()
    
        while unparsed:
            token = unparsed.pop().strip()
    
            parts = token.partition('=')
            if parts[-1] != '':  # parsing for --option=value type input
                token = parts[0]
    
            allFlags = root.localFlags + self.globalFlags
            found = False
    
            for flag in allFlags:
                if flag.name == token:
                    logger.debug("matched token: %s with flag: %s", token, flag.name)
                    parsed.append(token)
                    found = True
                    break
    
            if not found:
                logger.debug("no flags match for token: %s", token)
                unparsed.append(token)
                break
    
        for flag in allFlags:
            suggestions[flag.name] = flag.helptext
    
        logger.debug("options parsed: %s, remaining unparsed: %s", parsed, unparsed)
        return parsed, unparsed, suggestions
    



from __future__ import print_function, absolute_import, unicode_literals

from pygments.token import Token
from pygments.util import ClassNotFound
from pygments.styles import get_style_by_name
from prompt_toolkit.styles import Style

styles_dict = {
        "": "ansicyan",
        "oci": "#C74634",
        "bottom-toolbar": "noreverse",
        "bottom-toolbar.text": "#888888 bg:default noreverse noitalic nounderline noblink",
        "bottom-toolbar.error": "fg:ansired",
        "required-parameter": "fg:ansired"
}



import os
import sys
import logging
import traceback
import logging.config

if not os.path.exists(os.path.expanduser("~/.goat/shell")):
    try:
        os.makedirs(os.path.expanduser("~/.goat/shell"))
    except OSError:
        print("failed to make config dirs for goat-shell")
        traceback.print_exc(file=sys.stdout)

logfile = os.path.expanduser("~/.goat/shell/error.log")
loggingConf = {
    "version": 1,
    "formatters": {
        "default": {
            "format": "%(asctime)-15s [%(levelname)-4s] %(name)s %(funcName)s:%(lineno)s - %(message)s",
        }
    },
    "handlers": {
        "null": {
            "class": "logging.NullHandler",
            "level": "ERROR"
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "INFO",
            "formatter": "default",
            "filename": logfile,
            "backupCount": 3,
            "maxBytes": 10485760  # 10MB
        }
    },
    "loggers": {
        "": {
            "level": "ERROR",
            "handlers": ["file"],
        },
        "urllib3": {
            "level": "ERROR",
            "handlers": ["file"],
            "propagate": False
        },
        "goatshell": {
            "level": "INFO",
            "handlers": ["file"],
            "propagate": False
        }
    },
}
logging.config.dictConfig(loggingConf)



from __future__ import absolute_import, unicode_literals, print_function
from subprocess import check_output
from prompt_toolkit.completion import Completer, Completion
from fuzzyfinder import fuzzyfinder
import logging
import shlex
import json
import os
import os.path
import asyncio
from . import misc

from goatshell.parser import Parser  # Replace this with your actual Parser import

logger = logging.getLogger(__name__)

class GoatCompleter(Completer):
    def __init__(self, parser):
        self.inline_help = True
        self.min_input_len = 0
        self.parser = parser
        self.current_json = None  # NEW: attribute to store the current JSON
        self.goat_dict = None

        # Dictionary mapping command tokens to their descriptions
        self.command_descriptions = {
            "aliyun": "Alibaba Cloud Command Line Interface",
            "aws": "Amazon Web Services",
            "az": "Microsoft Azure Cloud Platform",
            "gcloud": "Google Cloud Platform",
            "goat": "Local Goat Application",
            "oci": "Oracle Cloud Infrastructure",
            "ibmcloud": "IBM Cloud Infrastructure"
            # Add more commands and descriptions as needed
        }

    def load_json(self, provider):
        DATA_DIR = os.path.dirname(os.path.realpath(__file__))
        DATA_PATH = os.path.join(DATA_DIR, f'data/{provider}.json')

        try:
            with open(DATA_PATH) as json_file:
                self.goat_dict = json.load(json_file)

            self.parser = Parser(DATA_PATH, provider)
        except Exception as ex:
            logger.error(f"Exception while loading JSON for {provider}: {ex}")

    def get_completions(self, document, complete_event, smart_completion=True):
        tokens = shlex.split(document.text_before_cursor.strip())

        if not tokens:
            # Check if a command is available before suggesting it
            available_commands = ["aliyun", "aws", "az", "gcloud", "goat", "oci", "ibmcloud"]
            for command in available_commands:
                if misc.is_command_available(command):
                    yield Completion(command, display=command, display_meta=self.command_descriptions.get(command, ""))
            return

        first_token = tokens[0]

        if first_token in ["oci", "aws", "gcloud", "az", "goat", "aliyun", "ibmcloud"] and self.current_json != first_token:
            self.load_json(first_token)
            self.current_json = first_token

        if len(tokens) == 1 and first_token in ["oci", "aws", "gcloud", "az", "goat", "aliyun", "ibmcloud"]:
            subcommands = self.parser.ast.children
            for subcmd in subcommands:
                yield Completion(subcmd.node, display=subcmd.node, display_meta=subcmd.help)
            return

        parsed, unparsed, suggestions = self.parser.parse_tokens(tokens)

        if suggestions is None:
            logger.error("suggestions are None")
            return

        if unparsed:
            last_token = unparsed[-1]
            if last_token.startswith("--") or last_token.startswith('-'):
                if not document.text_before_cursor.endswith("--") or not document.text_before_cursor.endswith('-'):
                    option_prefix = last_token
                    completions = fuzzyfinder(option_prefix, suggestions.keys())
                    for key in completions:
                        yield Completion(key, start_position=-len(option_prefix), display=key, display_meta=suggestions.get(key, ""))
            else:
                completions = fuzzyfinder(last_token, suggestions.keys())
                for key in completions:
                    yield Completion(key, start_position=-len(last_token), display=key, display_meta=suggestions.get(key, ""))
        else:
            for key in suggestions.keys():
                yield Completion(key, display=key, display_meta=suggestions.get(key, ""))

    async def get_completions_async(self, document, complete_event):
        logger.debug("Entering get_completions")
        completions = await asyncio.to_thread(self.get_completions, document, complete_event)
        for completion in completions:
            yield completion




from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit import Application, PromptSession
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import FileHistory
from prompt_toolkit.styles import Style
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

import time
import os
import sys
import subprocess
import re
import click
import logging
from . import misc as misc
from goatshell.style import styles_dict 
from goatshell.completer import GoatCompleter 
from goatshell.parser import Parser  
from goatshell.ui import getLayout  

logger = logging.getLogger(__name__)
current_service = 'oci'  # You can set it to 'aws' if you prefer AWS mode initially
os.environ['AWS_PAGER'] = ''

# Define key bindings class
class CustomKeyBindings(KeyBindings):
    def __init__(self, app, goatshell_instance):
        super().__init__()
        self.app = app  # Store the Application reference
        self.goatshell_instance = goatshell_instance

        @self.add(Keys.F12)
        def handle_f10(event):
            sys.exit()

        @self.add(Keys.F10)
        def handle_f12(event):
            self.profile = self.goatshell_instance.get_profile(self.goatshell_instance.prefix)  # Access the get_profile method of Goatshell
            getLayout()
            self.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to reprompt.

        @self.add(Keys.F8)
        def handle_f8(event):
            getLayout()
            self.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to reprompt.

# Define the main Goatshell class
class Goatshell(object):
    def __init__(self, app, completer, parser):
        getLayout()
        self.style = Style.from_dict(styles_dict)
        self.app = app
        self.completer = completer
        self.prefix = 'oci'
        self.current_input = None
        self.aws_profiles = misc.read_aws_profiles()
        self.oci_profiles = misc.read_oci_profiles()
        self.aws_index = 0
        self.oci_index = 0
        self.profile = 'DEFAULT' # init this variable before we call the get_profile function
        self.profile = self.get_profile(self.prefix)
        self.key_bindings = CustomKeyBindings(self.app, self)
        shell_dir = os.path.expanduser("~/goat/shell/")
        if not os.path.exists(shell_dir):
            os.makedirs(shell_dir)
        try:
            self.history = FileHistory(os.path.join(shell_dir, "history"))
        except:
            self.history = InMemoryHistory()

        self.session = PromptSession(history=self.history, auto_suggest=AutoSuggestFromHistory(),
                                     completer=self.completer, complete_while_typing=True,
                                     enable_history_search=True, vi_mode=True,
                                     key_bindings=self.key_bindings)

        root_name, json_path = self.get_service_info()
        try:
            self.parser = Parser(json_path, root_name)
        except Exception as e:
            logger.info(f"Exception caught: {e}")
            logger.info(f"json_path = {json_path}, root_name = {root_name}")
   
    def get_account_or_tenancy(self, profile):
        if self.prefix == 'oci':
            # label as account for variable consistency and get the ocid
            account = misc.get_oci_tenant(profile_name=profile).id
        elif self.prefix == 'aws':
            account = misc.get_aws_account(profile_name=profile)
        else:
            return 'UNKNOWN'
        return account

    def get_profile(self, mode):
        if mode == 'aws':
            if self.aws_profiles:
                if self.profile in self.aws_profiles:
                    self.aws_index = self.aws_profiles.index(self.profile)
                self.aws_index = (self.aws_index + 1) % len(self.aws_profiles)
                self.profile = self.aws_profiles[self.aws_index]
            else:
                self.profile = 'DEFAULT'
        elif mode == 'oci':
            if self.oci_profiles:
                if self.profile in self.oci_profiles:
                    self.oci_index = self.oci_profiles.index(self.profile)
                self.oci_index = (self.oci_index + 1) % len(self.oci_profiles)
                self.profile = self.oci_profiles[self.oci_index]
            else:
                self.profile = 'DEFAULT'
        else:
            self.profile = 'DEFAULT'
        return self.profile

    def get_service_info(self):
        json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), f'data/{self.prefix}.json')
        global current_service  # Declare current_service as a global variable
        if current_service == 'aws':
            return 'aws', json_path
        elif current_service == 'oci':
            return 'oci', json_path
        elif current_service == 'gcloud':
            return 'gcloud', json_path
        elif current_service == 'az':
            return 'az', json_path
        elif current_service == 'goat':
            return 'goat', json_path
        elif current_service == 'aliyun':
            return 'aliyun', json_path
        elif current_service == 'ibmcloud':
            return 'ibmcloud', json_path
        else:
            return 'aws', json_path
    
    def create_toolbar(self):
        self.upper_profile = self.profile.upper()
        self.upper_prefix = self.prefix.upper()
        return HTML(
            f'Current Cloud: <u>{self.upper_prefix}</u>  <b>F8</b> Usage <b>F10</b> Toggle Profile: <u>{self.upper_profile}</u> <b>F12</b> Quit'
        )

    def set_parser_and_completer(self, api_type):
        self.prefix = api_type.lower()  # Set prefix
        json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), f'data/{api_type}.json')
        self.parser = Parser(json_path, api_type)  # Reset parser
        self.completer = GoatCompleter(self.parser)  # Reset completer
        self.session.completer = self.completer  # Reset session completer

    def run_cli(self):
        global current_service
        logger.info("running goat event loop")
        while True:
            try:
                user_input = self.session.prompt(f'goat> ',
                                                 style=self.style,
                                                 enable_history_search=True,
                                                 vi_mode=True,
                                                 complete_while_typing=True,
                                                 bottom_toolbar=self.create_toolbar())
            except (EOFError, KeyboardInterrupt):
                sys.exit()

            if user_input == 're-prompt':
                user_input = ''
                continue
            api_type = user_input.split(' ')[0]

            if api_type.lower() != self.prefix:  # If a different prefix is detected
                if api_type.lower() in ['oci', 'aws', 'gcloud', 'az', 'goat', 'aliyun', 'ibmcloud']:
                    self.set_parser_and_completer(api_type.lower())

            user_input = re.sub(r'[-]{3,}', '--', user_input)
            if user_input == "clear" or user_input == 'c':
                user_input = 'clear'
                click.clear()
            elif user_input == "exit" or user_input == 'e':
                user_input = 'exit'
                sys.exit()
            elif user_input == "help" or user_input == 'h':
                getLayout()
                continue

            if user_input.startswith("!"):
                user_input = user_input[1:]
            else:
                if user_input:
                    last_token = user_input.split(' ')[-1]
                    first_token = user_input.split(' ')[0]
                    # check for aliyun and make sure its in our path
                    if first_token.lower() == 'aliyun':
                        if not misc.is_command_available(first_token.lower()):
                            print(f"INFO: {first_token.lower()} not in path for execution")
                            user_input = ''

                    if first_token.lower() == 'oci':
                        if self.profile not in self.oci_profiles:
                            # need to verify our current profile exists
                            self.profile = self.get_profile(first_token.lower())
                        if '--compartment-id' in user_input or '--tenancy-id' in user_input and 'ocid' not in user_input:
                            try:
                                OCID = self.get_account_or_tenancy(self.profile)
                                user_input += f' {OCID}'
                            except:
                                pass
                        if '--user-id' in user_input and 'ocid' not in user_input:
                            try:
                                OCID = misc.get_oci_user(self.profile)
                                user_input += f' {OCID}'
                            except:
                                pass
                    if first_token.lower() == 'aws':
                        if self.profile not in self.aws_profiles:
                            # need to verify our current profile exists
                            self.profile = self.get_profile(first_token.lower())
                        if last_token == '--user-name':
                            try:
                                USER = misc.get_aws_user(self.profile)
                                user_input += f' {USER}'
                            except:
                                pass
                    if first_token.lower() not in ['gcloud','az','goat','aliyun'] and '--profile' not in user_input:
                        user_input = user_input + ' --profile ' + self.profile
                    if '-o' in user_input and 'json' in user_input:
                        user_input += ' | pygmentize -l json'
            p = subprocess.Popen(user_input, shell=True)
            p.communicate()

